import React, { LegacyRef, useEffect, useRef, useState } from 'react';
import * as d3 from 'd3';
import { PlotTooltip } from '@graphpolaris/shared/lib/components/tooltip';

export type BarPlotProps = {
  data: { category: string; count: number }[];
  numBins: number;
  typeBarPlot: 'numerical' | 'categorical';
};

export const BarPlot = ({ typeBarPlot: typeBarplot, numBins, data }: BarPlotProps) => {
  const svgRef = useRef<SVGSVGElement | null>(null);
  const [dimensions, setDimensions] = useState({ width: 0, height: 0 });
  const [tooltipData, setTooltipData] = useState<{ x: number; y: number; content: React.ReactNode } | null>(null);

  useEffect(() => {
    if (!svgRef.current) return;

    const widthSVG: number = +svgRef.current.clientWidth;
    const heightSVG: number = +svgRef.current.clientHeight;

    setDimensions({ width: widthSVG, height: heightSVG });
    const svgPlot = d3.select(svgRef.current);

    svgPlot.selectAll('*').remove();

    let groupMargin: d3.Selection<SVGGElement, unknown, null, undefined>;
    let widthSVGwithinMargin: number = 0.0;
    let heightSVGwithinMargin: number = 0.0;

    // Barplot for categorical data
    if (typeBarplot == 'categorical') {
      let marginPercentage = {
        top: 0.19,
        right: 0.02,
        bottom: 0.19,
        left: 0.19,
      };
      let margin = {
        top: marginPercentage.top * heightSVG,
        right: marginPercentage.right * widthSVG,
        bottom: marginPercentage.bottom * heightSVG,
        left: marginPercentage.left * widthSVG,
      };

      groupMargin = svgPlot.append('g').attr('transform', `translate(${margin.left},${margin.top})`);
      widthSVGwithinMargin = widthSVG - margin.left - margin.right;
      heightSVGwithinMargin = heightSVG - margin.top - margin.bottom;

      // Data processing
      const dataFiltered = data.filter((item) => item.category !== undefined);
      const dataSorted = dataFiltered.sort((a, b) => b.count - a.count);
      const maxCount = d3.max(dataSorted, (d) => d.count) ?? 1;
      const xScale = d3
        .scaleBand()
        .range([0, widthSVGwithinMargin])
        .domain(
          dataSorted.map(function (d) {
            return d.category;
          })
        )
        .padding(0.2);

      // send this
      const yScale = d3.scaleLinear().domain([0, maxCount]).range([heightSVGwithinMargin, 0]);

      // here out to axis component
      const yAxis1 = d3.axisLeft(yScale).tickValues([0]).tickFormat(d3.format('d')); // to show 0 without decimanls
      let yAxis2: d3.Axis<d3.NumberValue>;

      if (maxCount < 10) {
        yAxis2 = d3.axisLeft(yScale).tickValues([maxCount]).tickFormat(d3.format('d'));
      } else {
        yAxis2 = d3.axisLeft(yScale).tickValues([maxCount]).tickFormat(d3.format('.2s'));
      }

      groupMargin
        .selectAll<SVGRectElement, d3.Bin<string, number>>('barplotCats')
        .data(dataSorted)
        .enter()
        .append('rect')
        .attr('x', (d) => xScale(d.category) || 0)
        .attr('y', (d) => yScale(d.count))
        .attr('width', xScale.bandwidth())
        .attr('height', (d) => heightSVGwithinMargin - yScale(d.count))
        .attr('fill', 'hsl(var(--clr-sec--400))')
        .attr('stroke', 'none')
        .on('mouseover', (event, d) => {
          const tooltipContent = (
            <div>
              <span>{d.category.toString()}</span>: <span>{d.count}</span>
            </div>
          );

          setTooltipData({ x: event.pageX + 10, y: event.pageY - 28, content: tooltipContent });
        })
        .on('mousemove', (event, d) => {
          if (tooltipData) {
            setTimeout(() => {
              setTooltipData((prevTooltipData) => {
                if (prevTooltipData) {
                  return {
                    ...prevTooltipData,
                    x: event.pageX + 10,
                    y: event.pageY - 28,
                  };
                }
                return prevTooltipData;
              });
            }, 16);
          }
        })
        .on('mouseleave', (d) => {
          setTooltipData(null);
        });

      groupMargin.append('g').call(yAxis1);
      groupMargin.append('g').call(yAxis2);

      // Barplot for numerical data
    } else {
      let marginPercentage = {
        top: 0.19,
        right: 0.17,
        bottom: 0.19,
        left: 0.4,
      };
      let margin = {
        top: marginPercentage.top * heightSVG,
        right: marginPercentage.right * widthSVG,
        bottom: marginPercentage.bottom * heightSVG,
        left: marginPercentage.left * widthSVG,
      };

      groupMargin = svgPlot.append('g').attr('transform', `translate(${margin.left},${margin.top})`);
      widthSVGwithinMargin = widthSVG - margin.left - margin.right;
      heightSVGwithinMargin = heightSVG - margin.top - margin.bottom;
      //console.log(typeBarplot, data);
      const dataCount = data.map((obj) => obj.count);

      const extentData = d3.extent(dataCount);
      const [min, max] = extentData as [number, number];

      const xScale = d3.scaleLinear().range([0, widthSVGwithinMargin]).domain([min, max]); //.nice();

      // d3.histogram -> deprecated
      // On d3.bin(), .thresholds(x.ticks(numBins)). Creates artifacts: not full rects on the plot
      // CHECK THIS: https://dev.to/kevinlien/d3-histograms-and-fixing-the-bin-problem-4ac5

      const rangeTH: number[] = d3.range(min, max, (max - min) / numBins);

      if (rangeTH.length != numBins) {
        rangeTH.pop();
      }

      const histogram = d3
        .bin()
        .value(function (d) {
          return d;
        })
        .domain([min, max])
        .thresholds(rangeTH);

      const bins = histogram(dataCount);
      //console.log(bins);
      const extentBins: [number, number] = d3.extent(bins, (d) => d.length) as [number, number];

      const yScale = d3
        .scaleLinear()
        .range([heightSVGwithinMargin, 0])
        //.domain(d3.extent(bins, (d) => d.length) as [number, number]);
        .domain([0, extentBins[1]] as [number, number]);

      groupMargin
        .selectAll<SVGRectElement, d3.Bin<number, number>>('barplotbins')
        .data(bins)
        .enter()
        .append('rect')
        .attr('x', 0)
        .attr('stroke', 'none')
        .attr('fill', 'hsl(var(--clr-sec--400))')
        .attr('transform', (d) => 'translate(' + xScale(d.x0 || 0) + ',' + yScale(d.length) + ')')
        .attr('width', (d) => xScale(d.x1 || 0) - xScale(d.x0 || 0))
        .attr('height', (d) => heightSVGwithinMargin - yScale(d.length));

      const xAxis = d3
        .axisBottom(xScale)
        //.tickValues([Math.round((min + max) / 2.0), max])
        .tickValues([max])
        .tickFormat(d3.format('.2s'));

      let xAxis2: d3.Axis<d3.NumberValue>;

      if (min < 10) {
        xAxis2 = d3.axisBottom(xScale).tickValues([min]).tickFormat(d3.format('d'));
      } else {
        xAxis2 = d3.axisBottom(xScale).tickValues([min]).tickFormat(d3.format('.2s'));
      }

      groupMargin
        .append('g')
        .attr('transform', 'translate(0,' + heightSVGwithinMargin + ')')
        .call(xAxis2);

      groupMargin
        .append('g')
        .attr('transform', 'translate(0,' + heightSVGwithinMargin + ')')
        .call(xAxis);

      const yAxis1 = d3.axisLeft(yScale).tickValues([0]).tickFormat(d3.format('d')); // to show 0 without decimanls
      //const yAxis2 = d3.axisLeft(yScale).tickValues([extentBins[1]]).tickFormat(d3.format('.2s'));

      let yAxis2: d3.Axis<d3.NumberValue>;

      if (extentBins[1] < 10) {
        yAxis2 = d3.axisLeft(yScale).tickValues([extentBins[1]]).tickFormat(d3.format('d'));
      } else {
        yAxis2 = d3.axisLeft(yScale).tickValues([extentBins[1]]).tickFormat(d3.format('.2s'));
      }
      // two axis for each number, it solves the 0.0 problem with ".2s" notatation
      groupMargin.append('g').call(yAxis1);
      groupMargin.append('g').call(yAxis2);
    }

    svgPlot.selectAll('.domain').style('stroke', 'hsl(var(--clr-sec--400))');
    svgPlot.selectAll('.tick line').style('stroke', 'hsl(var(--clr-sec--400))');
    svgPlot.selectAll('.tick text').attr('class', 'font-mono').style('stroke', 'none').style('fill', 'hsl(var(--clr-sec--500))');

    groupMargin
      .append('rect')
      .attr('x', 0.0)
      .attr('y', 0.0)
      .attr('width', widthSVGwithinMargin)
      .attr('height', heightSVGwithinMargin)
      .attr('rx', 0)
      .attr('ry', 0)
      .attr('stroke-width', 1)
      .attr('fill', 'none')
      .attr('stroke', 'hsl(var(--clr-sec--400))');
  }, [data]);

  return (
    <div className="svg">
      <svg
        ref={svgRef}
        className="container"
        width="100%"
        height="100%"
        //preserveAspectRatio="xMidYMid meet"
        //viewBox={`0 0 ${dimensions.width} ${dimensions.height}`}
      ></svg>
      {tooltipData && <PlotTooltip x={tooltipData.x} y={tooltipData.y} content={tooltipData.content} />}
    </div>
  );
};

export default BarPlot;
