import React, { useEffect, useRef, useState } from 'react';
import {
  select,
  selectAll,
  scaleLinear,
  CurveFactory,
  axisBottom,
  axisLeft,
  area,
  line,
  curveCardinal,
  curveStepAfter,
  curveNatural,
  curveLinear,
  min,
  max,
} from 'd3';
import { visualizationColors } from 'config';

// Define the LineChartProps type
export type LineChartProps = {
  data: { x: number; y: number }[][]; // Array of arrays of data points
  xAxisLabel: string;
  yAxisLabel: string;
  showMarks: boolean;
  curveInterpolation: string[];
  isArea: boolean;
};

// LineChart component
const LineChart: React.FC<LineChartProps> = ({ data, xAxisLabel, yAxisLabel, showMarks, curveInterpolation, isArea }) => {
  // Ref for SVG element
  const svgRef = useRef<SVGSVGElement | null>(null);
  // State for dimensions of the SVG container
  const [dimensions, setDimensions] = useState({ width: 0, height: 0 });
  // State to track mark visibility

  // Effect to update dimensions when the component mounts or data changes
  useEffect(() => {
    if (!svgRef.current) return;

    // Get width and height of the SVG container
    const widthSVG: number = +svgRef.current.clientWidth;
    const heightSVG: number = +svgRef.current.clientHeight;

    // Update dimensions state
    setDimensions({ width: widthSVG, height: heightSVG });

    // Define margin
    const marginPercentage = {
      top: 0.245,
      right: 0.245,
      bottom: 0.245,
      left: 0.245,
    };
    const margin = {
      top: marginPercentage.top * heightSVG,
      right: marginPercentage.right * heightSVG,
      bottom: marginPercentage.bottom * heightSVG,
      left: marginPercentage.left * heightSVG,
    };

    // Define innerwidth and innerheight
    const innerWidth = widthSVG - margin.left - margin.right;
    const innerHeight = heightSVG - margin.top - margin.bottom;

    // Remove existing elements from SVG
    select(svgRef.current).selectAll('*').remove();

    // Create SVG container
    const svg = select(svgRef.current).attr('width', widthSVG).attr('height', heightSVG);

    // Create group for plotting area
    const plotArea = svg.append('g').attr('transform', `translate(${margin.left},${margin.top})`);

    // Define scales
    const xScale = scaleLinear()
      .domain([min(data.flatMap((d) => d.map((p) => p.x))) || 0, max(data.flatMap((d) => d.map((p) => p.x))) || 0])
      .range([0, innerWidth]);

    const yScale = scaleLinear()
      .domain([min(data.flatMap((d) => d.map((p) => p.y))) || 0, max(data.flatMap((d) => d.map((p) => p.y))) || 0])
      .range([innerHeight, 0]);

    // Add grid lines for x axis
    plotArea
      .append('g')
      .attr('class', 'grid')
      .attr('transform', `translate(0, ${innerHeight})`)
      .call(
        axisBottom(xScale)
          .tickSize(-innerHeight)
          .tickFormat(() => '')
          .tickSizeOuter(0)
          .ticks(xScale.ticks().length * 2),
      );

    // Add grid lines for y axis
    plotArea
      .append('g')
      .attr('class', 'grid')
      .call(
        axisLeft(yScale)
          .tickSize(-innerWidth)
          .tickFormat(() => '')
          .tickSizeOuter(0)
          .ticks(yScale.ticks().length / 2),
      );

    // CSS styling for grid lines
    selectAll('.grid line').style('stroke', visualizationColors.GPNeutral.colors[14][3]);

    // Calculate suitable tick values
    const xTicks = xScale.ticks(); // Use domain values for x-axis
    const yTicks = yScale.ticks(); // Use domain values for y-axis

    // Draw axes
    const xAxis = axisBottom(xScale).tickValues(xTicks).tickSize(0).tickPadding(8).tickSizeOuter(4);
    const yAxis = axisLeft(yScale).tickValues(yTicks).tickSize(0).tickPadding(8).tickSizeOuter(4);

    const AxisTextCol = visualizationColors.GPNeutral.colors[14];
    plotArea
      .append('g')
      .attr('class', 'x-axis')
      .attr('transform', `translate(0, ${innerHeight})`)
      .call(xAxis)
      .selectAll('.tick text')
      .attr('class', 'font-mono')
      .attr('fill', AxisTextCol[13])
      .style('stroke', 'none');

    plotArea
      .append('g')
      .attr('class', 'y-axis')
      .call(yAxis)
      .selectAll('.tick text')
      .attr('class', 'font-mono')
      .attr('fill', AxisTextCol[13])
      .style('stroke', 'none');

    // Add x-axis label
    plotArea
      .append('text')
      .attr('class', 'x-axis-label font-Inter font-bold text-xs')
      .attr('x', innerWidth / 2)
      .attr('y', innerHeight + 38)
      .attr('text-anchor', 'middle')
      .text(xAxisLabel);

    // Add y-axis label
    plotArea
      .append('text')
      .attr('class', 'y-axis-label font-Inter font-bold text-xs')
      .attr('transform', `rotate(-90)`)
      .attr('x', -innerHeight / 2)
      .attr('y', -38)
      .attr('text-anchor', 'middle')
      .text(yAxisLabel);

    // Specify colors to use
    const catColors = visualizationColors.GPCat.colors[14];

    let d3area = area<{ x: number; y: number }>()
      .x((d) => xScale(d.x))
      .y0(innerHeight)
      .y1((d) => yScale(d.y));

    let d3line = line<{ x: number; y: number }>()
      .x((d) => xScale(d.x))
      .y((d) => yScale(d.y));

    const interpolationMap: Record<string, CurveFactory> = {
      smooth: curveCardinal,
      step: curveStepAfter,
      natural: curveNatural,
    };

    if (Array.isArray(curveInterpolation)) {
      curveInterpolation.forEach((option) => {
        const curveFunction = interpolationMap[option];
        if (curveFunction) {
          d3line = d3line.curve(curveFunction);
          d3area = d3area.curve(curveFunction);
        } else {
          d3line = d3line.curve(curveLinear);
          d3area = d3area.curve(curveLinear);
        }
      });
    } else {
      const curveFunction = interpolationMap[curveInterpolation];
      if (curveFunction) {
        d3line = d3line.curve(curveFunction);
        d3area = d3area.curve(curveFunction);
      } else {
        d3line = d3line.curve(curveLinear);
        d3area = d3area.curve(curveLinear);
      }
    }

    // Draw area under the curve for each line of data
    if (isArea) {
      data.forEach((lineData, index) => {
        const colorIndex = index % catColors.length;
        plotArea.append('path').datum(lineData).attr('fill', catColors[colorIndex]).style('opacity', 0.3).attr('d', d3area);
      });
    }

    // Draw lines
    data.forEach((lineData, index) => {
      const colorIndex = index % catColors.length;
      plotArea
        .append('path')
        .datum(lineData)
        .attr('fill', 'none')
        .attr('stroke', catColors[colorIndex])
        .attr('stroke-width', 1.5)
        .attr('d', d3line);

      // Add marks (circles) for each data point
      if (showMarks) {
        // Add marks if showMarks is true
        plotArea
          .selectAll(`.mark-${index}`)
          .data(lineData)
          .enter()
          .append('circle')
          .attr('class', `mark mark-${index}`)
          .attr('cx', (d) => xScale(d.x))
          .attr('cy', (d) => yScale(d.y))
          .attr('r', 4.5)
          .attr('fill', 'white')
          .attr('stroke', catColors[colorIndex])
          .attr('stroke-width', 1.5);
      }
    });
  }, [data, dimensions.height, dimensions.width, showMarks, curveInterpolation, isArea]);

  // Return the LineChart component
  return (
    <svg
      ref={svgRef}
      className="container"
      width="100%"
      height="100%"
      preserveAspectRatio="xMidYMid meet"
      viewBox={`0 0 ${dimensions.width} ${dimensions.height}`}
    ></svg>
  );
};

export default LineChart;
